#include "testApp.h"
#include <string>
#include <sstream>
#include <iostream>
#include "ofxFileDialog.h"
#include "ofUtils.h"
#include "ofCamera.h"
#include "ofxSimpleButton.h"
#include "keyPresses.hpp"
#include "selection.hpp"
#include "loadingWorld.hpp"
#include "mouse.h"
#include "drawFunction.h"
#include "b2dJson.h"

using namespace std;

void testApp::playButtonFunction(){
    box2d.resetTimeStep();
    if (paused){
        paused = false;
    } else {
        paused = false;
        keyPressed('p');
    }
}

void testApp::pauseButtonFunction(){
    if (circles.size() > 0){
        paused = true;
        box2d.setTimeStep(0);
    }
}

void testApp::stopButtonFunction(){
    paused = false;
    circles.clear();
}

void testApp::recButtonFunction(){
    currentMode = RECTANGLE;
}

void testApp::lineButtonFunction(){
    currentMode = LINE;
}

void testApp::circleButtonFunction(){
    currentMode = STREAM;
}


//--------------------------------------------------------------
void testApp::setup() {
    
    // BUTTONS
    buttons[0].setup(this, &testApp::playButtonFunction, 480, 10, 50, 50, "img/grayPlayButton");
    buttons[1].setup(this, &testApp::pauseButtonFunction, 520, 10, 50, 50, "img/pauseButton");
    buttons[2].setup(this, &testApp::stopButtonFunction, 560, 10, 50, 50, "img/grayStopButton");
    buttons[3].setup(this, &testApp::recButtonFunction, 60, 70, 50, 50, "img/rectangle");
    buttons[4].setup(this, &testApp::lineButtonFunction, 60, 120, 50, 50, "img/line");
    
    // Setting up PLAY button/LOADING IMAGES
    // -----------------------------
    inPortal.load("img/blueWhirl.png");
    outPortal.load("img/orangeWhirl.png");
    mouseHand.load("img/mouse_hand.png");
    clenchedMouse.load("img/clenched-fist.png");
    flagImage.load("img/flag.png");
    buttonOutline.load("img/square_outline.png");
    rectangleIcon.load("img/rectangle.png");
    trampolineIcon.load("img/yellowRec.png");
    lineIcon.load("img/line.png");
    circleIcon.load("img/circle.png");
    
    // 2D DRAWING WORLD SETUP
    // ------------------------------
	ofSetVerticalSync(true);
	ofBackgroundHex(0xfdefc2);

	box2d.init();
    box2d.enableEvents();   // <-- turn on the event listener
	box2d.setGravity(0, 5); // turned it negative after camera thing
	box2d.setFPS(60.0);
	box2d.registerGrabbing();
	
	ofAddListener(box2d.contactStartEvents, this, &testApp::contactStart);
	ofAddListener(box2d.contactEndEvents, this, &testApp::contactEnd);
    
    selected.ptr = NULL;
    selected.mode = SELECTION; //signifies nothing selected
    selected.coordinates = b2Vec2(0, 0);

    // LOAD PIANO SOUND FILES
    for (int i = 1; i < 64; i++){
        string fileName = "piano/" + to_string(i) + ".wav";
        sound[i].loadSound(fileName);
        sound[i].setMultiPlay(true);
        sound[i].setLoop(false);
    }
    currentSound = 28;
    
    
    // FIRST STREAM
    // ------------------------
    streams.push_back(b2Vec2(SCREEN_WIDTH / 2, CIRCLE_RADIUS));
    
    // Setup Camera
    // -------------------------
    myCamera = ofCamera();
    myCamera.enableOrtho();
    myCamera.setPosition(SCREEN_WIDTH / 2, SCREEN_HEIGHT / 2, 10);
    myCamera.lookAt(ofVec3f(SCREEN_WIDTH / 2, SCREEN_HEIGHT / 2, 0));
    myCamera.setScale(1, -1, 1);
}


//--------------------------------------------------------------
void testApp::contactStart(ofxBox2dContactArgs &e) {
	if(e.a != NULL && e.b != NULL) {
        
        if((e.a->GetType() == b2Shape::e_circle && e.b->GetType() == b2Shape::e_polygon)) {
            e.b->GetBody()->setHit();
            int rectSound = e.b->GetBody()->getSoundFile();
            sound[rectSound].play();
            if (e.b->GetBody()->getAcceleration() != 0){
                b2Vec2 currentVel = e.a->GetBody()->GetLinearVelocity();
                e.a->GetBody()->SetLinearVelocity(b2Vec2(currentVel.x, currentVel.y + e.b->GetBody()->getAcceleration()));
            }
        }
        
        if((e.a->GetType() == b2Shape::e_polygon && e.b->GetType() == b2Shape::e_circle)) {
            e.a->GetBody()->setHit();
            int rectSound = e.a->GetBody()->getSoundFile();
            sound[rectSound].play();
            
            if (e.a->GetBody()->getAcceleration() != 0){
                b2Vec2 currentVel = e.b->GetBody()->GetLinearVelocity();
                e.b->GetBody()->SetLinearVelocity(b2Vec2(currentVel.x, currentVel.y + e.a->GetBody()->getAcceleration()));
            }
        }
	}
}

//--------------------------------------------------------------
void testApp::contactEnd(ofxBox2dContactArgs &e) {
	if(e.a != NULL && e.b != NULL) {
        e.a->GetBody()->unsetHit();
        e.b->GetBody()->unsetHit();
	}
}



//--------------------------------------------------------------
void testApp::update() {
//    deleteOffScreen();
    
//    if (checkPoint){
//        b2dJson json;
//        string mystring = json.writeToString(box2d.world);
//        
//        
//        string errorMsg;
//        b2World* myWorld = json.readFromString(mystring, errorMsg);
//    }
    
//    video.update();
    
	box2d.update();
    
    if (trickle){
        if(trickleCounter % TRICKLE_FREQUENCY == 0){
            deleteOffScreen();
            for (int i = 0; i < streams.size(); i++){
                b2Vec2 coordinates = streams[i];
                // Create new circle every trickleFrequency
                shared_ptr <ofxBox2dCircle> c = shared_ptr <ofxBox2dCircle>(new ofxBox2dCircle);
                c.get()->setPhysics(1, 0.5, 0.9);
                c.get()->setup(box2d.getWorld(), coordinates.x, coordinates.y, CIRCLE_RADIUS);
                c.get()->setCircleID(circleIDCounter);
                circleIDCounter++;
                circles.push_back(c);
                vector<b2Vec2> newPath;
                ballTrajectories.push_back(newPath);
            }
        }
    }
    trickleCounter++;
    
    
    // ROTATE ALL PLATFORMS
    for (platform p : trampolines){
        if (p.rotating){
            b2Transform transform = p.ptr->body->GetTransform();
            b2Vec2 position = transform.p;
            float angle = transform.q.GetAngle();
            angle = angle + 0.01;
            p.ptr->body->SetTransform(p.ptr->body->GetPosition(), angle);
        }
    }
    for (platform p : rectangles){
        if (p.rotating){
            b2Transform transform = p.ptr->body->GetTransform();
            b2Vec2 position = transform.p;
            float angle = transform.q.GetAngle();
            angle = angle + 0.01;
            p.ptr->body->SetTransform(p.ptr->body->GetPosition(), angle);
        }
    }
    
    // SAVE DATA FOR BALL TRAJECTORIES
    // put into ball trajectories data structure
    for(size_t i=0; i<circles.size(); i++) {
        int circleID = circles[i].get()->getCircleID();
        ofVec2f pos = circles[i].get()->getPosition();
        while ((ballTrajectories.size() - 1) < circleID){
            vector<b2Vec2> newPath;
            ballTrajectories.push_back(newPath);
        }
        ballTrajectories[circleID].push_back(b2Vec2(pos.x, pos.y));
    }
}

bool testApp::intersection(b2Vec2 portalCenter, ofVec2f circleCenter){
    return (pow(portalCenter.x - circleCenter.x, 2) + pow(portalCenter.y - circleCenter.y, 2) <= pow(PORTAL_RADIUS, 2));
}

bool testApp::checkTeleport(int circleIndex){
    bool teleportDetected = false;
    for (int i = 0; i < inPortals.size(); i++){
        if (intersection(inPortals[i].coordinates, circles[circleIndex].get()->getPosition())){
            teleportDetected = true;
            // if found within an In-Portal...
            int pairNumber = inPortals[i].pairNumber;
            ofVec2f velocity = circles[circleIndex].get()->getVelocity();
            circles.erase(circles.begin() + circleIndex);
            
            // check each Out-Portal that matches and create a new circle
            for (int j = 0; j < outPortals.size(); j++){
                if (outPortals[j].pairNumber == pairNumber){
                    
                    shared_ptr <ofxBox2dCircle> c = shared_ptr <ofxBox2dCircle>(new ofxBox2dCircle);
                    c.get()->setPhysics(1, 0.5, 0.9);
                    c.get()->setup(box2d.getWorld(), outPortals[j].coordinates.x, outPortals[j].coordinates.y, CIRCLE_RADIUS);
                    c.get()->setVelocity(velocity);
                    c.get()->setCircleID(circleIDCounter); // give new ID
                    circleIDCounter++;
                    circles.push_back(c);
                }
            }
        }
    }
    
    
    // also create a new vector for Ball Trajectory structure, but save the original vector
    //                    ballTrajectories.push_back(ballTrajectories[circleIndex]); // save original
    //                    vector <b2Vec2> newTrajectoriesPath;
    //                    ballTrajectories[circleIndex] = newTrajectoriesPath;
//    ballTrajectories[circleIndex].clear();
    
    return teleportDetected;
}


void testApp::deleteOffScreen(){
    ofVec3f test3 = ofVec3f(myCamera.getPosition().x / OFX_BOX2D_SCALE, myCamera.getPosition().y / OFX_BOX2D_SCALE, 0);
    ofVec3f cScale = myCamera.getScale();
    float scaleAverage = cScale.x   ;
    float screenXOffset = ((SCREEN_WIDTH / OFX_BOX2D_SCALE) / 2) * scaleAverage;
    float screenYOffset = ((SCREEN_HEIGHT / OFX_BOX2D_SCALE) / 2) * scaleAverage;
//    cout << "scaleAverage:" << scaleAverage << endl;
//    cout << "xOffset:" << screenXOffset << ", yOffset:" << screenYOffset << endl;
//    float screenXOffset = ((SCREEN_WIDTH / OFX_BOX2D_SCALE) / 2) ;
//    float screenYOffset = ((SCREEN_HEIGHT / OFX_BOX2D_SCALE) / 2) ;
    b2Vec2 lower( test3.x - screenXOffset, test3.y - screenYOffset );
    b2Vec2 upper( test3.x + screenXOffset, test3.y + screenYOffset );
    b2AABB aabb;
    aabb.lowerBound = lower;
    aabb.upperBound = upper;
    vector< shared_ptr<ofxBox2dCircle> > toRemove;
    for (shared_ptr<ofxBox2dCircle> c : circles){
        b2Fixture* head = c.get()->body->GetFixtureList();
        for (b2Fixture* f = c.get()->body->GetFixtureList(); f; f = f->GetNext()){
            // if fixture's AABB b2TestOverlap overlaps with screen AABB
            if (!b2TestOverlap(f->GetAABB(0), aabb)){
                toRemove.push_back(c);
            }
        }
    }
    for (shared_ptr<ofxBox2dCircle> c : toRemove){
        circles.erase(std::remove(circles.begin(), circles.end(), c), circles.end());
    }
    
//    class MyQueryCallback : public b2QueryCallback {
//    public:
//        std::vector<b2Body*> foundBodies;
//
//        bool ReportFixture(b2Fixture* fixture) {
//            foundBodies.push_back( fixture->GetBody() );
//            return true;//keep going to find all fixtures in the query area
//        }
//    };
//    
//    ofVec3f test1 = ofVec3f(0, 0, 0);
//    ofVec3f test2 = ofVec3f(SCREEN_WIDTH / OFX_BOX2D_SCALE, SCREEN_HEIGHT / OFX_BOX2D_SCALE, 0);
//    ofVec3f test3 = ofVec3f(myCamera.getPosition().x / OFX_BOX2D_SCALE, myCamera.getPosition().y / OFX_BOX2D_SCALE, 0);
//    
//    float screenXOffset = (SCREEN_WIDTH / OFX_BOX2D_SCALE) / 2;
//    float screenYOffset = (SCREEN_HEIGHT / OFX_BOX2D_SCALE) / 2;
    
//    cout << "lower.x:" << test1.x << ", lower.y:" << test1.y << endl;
//    cout << "upper.x:" << test2.x << ", upper.y:" << test2.y << endl;
//    cout << "screen.x:" << test3.x << ", screen.y:" << test3.y << endl;
//    b2Vec2 lower( test3.x - screenXOffset, test3.y - screenYOffset );
//    b2Vec2 upper( test3.x + screenXOffset, test3.y + screenYOffset );
//    
    
//    MyQueryCallback queryCallback;
//    b2AABB aabb;
//    aabb.lowerBound = lower;
//    aabb.upperBound = upper;
//    box2d.getWorld()->QueryAABB( &queryCallback, aabb );
//    
//    //draw a point on each body in the query area
//    glPointSize(6);
//    glBegin(GL_POINTS);
//    for (int i = 0; i < queryCallback.foundBodies.size(); i++) {
//        b2Vec2 pos = queryCallback.foundBodies[i]->GetPosition();
//        glVertex2f( pos.x, pos.y );
//    }
//    glEnd();
}

//--------------------------------------------------------------
void testApp::draw() {
    drawFunction d = drawFunction(this);
    d.draw();
}

//--------------------------------------------------------------
void testApp::keyPressed(int key) {
    keyPresses k = keyPresses(this);
    k.keyPressed(key);
}

//--------------------------------------------------------------
void testApp::keyReleased(int key) {
    keyPresses k = keyPresses(this);
    k.keyReleased(key);
}

//--------------------------------------------------------------
void testApp::mouseMoved(int x, int y ) {
    mouse m = mouse(this);
    m.mouseMoved(x, y);
}

//--------------------------------------------------------------
void testApp::mouseDragged(int screen_x, int screen_y, int button) {
    mouse m = mouse(this);
    m.mouseDragged(screen_x, screen_y, button);
}

//--------------------------------------------------------------
void testApp::mousePressed(int screen_x, int screen_y, int button) {
    mouse m = mouse(this);
    m.mousePressed(screen_x, screen_y, button);
}

bool testApp::detectSelection(int x, int y){
    selection s = selection(this);
    return s.detectSelection(x, y);
}

//--------------------------------------------------------------
void testApp::mouseReleased(int screen_x, int screen_y, int button) {
    mouse m = mouse(this);
    m.mouseReleased(screen_x, screen_y, button);
}

//--------------------------------------------------------------
void testApp::resized(int w, int h){
}

//--------------------------------------------------------------
void testApp::loadWorld(string file) {
    loadingWorld l = loadingWorld(this);
    l.loadWorld(file);
}

void testApp::saveWorld() {
    loadingWorld l = loadingWorld(this);
    l.saveWorld();
}